/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/

#include "dp.h"
#include "cvtFast.h"


void freeCAresources(void);
void dpInitializeCA();
void dpTerminateCA();
void dpRegisterCA(void *dummy, int fd, int condition);

void processReadbackGrGetCallback(struct event_handler_args args);
void processReadbackMonitorEvent(struct event_handler_args args);
void processSetPointGrGetCallback(struct event_handler_args args);
void processSetPointMonitorEvent(struct event_handler_args args);
void processStatusGetCallback(struct event_handler_args args);
void processStatusMonitorEvent(struct event_handler_args args);

void updateStatistics(DpStruct *dpStruct);

static void updateOperatingRangeAndUnits(DpStruct *dpStruct);
static void caHeartBeat(XtPointer client_data, XtIntervalId *id);
static void dpUpdateDiff5(XtPointer client_data, XtIntervalId *id);
static void dpProcessCA(XtPointer client_data, int *source, XtInputId *id);

static XmString str;

void freeCAresources(void)
{
  DpStruct *ptr, *savedPtr;

  ptr = listHead;
  while (ptr != (DpStruct *)NULL) {
    if (ptr->setPointEvid != NULL) ca_clear_event(ptr->setPointEvid);
    if (ptr->readbackEvid != NULL) ca_clear_event(ptr->readbackEvid);
    if (ptr->setPointChid != NULL) ca_clear_channel(ptr->setPointChid);
    if (ptr->readbackChid != NULL) ca_clear_channel(ptr->readbackChid);
    savedPtr = ptr;
    ptr = ptr->next;
    free((char *)savedPtr);
  }
  listHead = listTail = (DpStruct *)NULL;
  numChannels = 0;

  ca_pend_event(CA_PEND_EVENT_TIME);
}


static void caHeartBeat(XtPointer client_data, XtIntervalId *id)
{
  ca_pend_event(CA_PEND_EVENT_TIME);    /* need this rather than ca_pend_io */
  /* reregister 10 second TimeOut function to handle lost connections, etc */
  XtAppAddTimeOut(appContext,10000,caHeartBeat,NULL);
}




/*
 * update fixed fields (operating range, units) in the display 
 */
static void updateOperatingRangeAndUnits(DpStruct *dpStruct)
{
  Dimension width;
  DpStruct *ele;
  Arg wargs[2];
  int nColumns;

  char loprString[24], hoprString[24];


/* if no set point defined for atomic dev, unmanage the set point text field
 *  and all those statistics fields which rely on a set point to have meaning */
  if (!dpStruct->foundSetPoint && dpStruct->entryType == AtomicDevice) {
	XtUnmanageChild(dpStruct->setPointTF);
	XtUnmanageChild(dpStruct->rmsDiffL);
	XtUnmanageChild(dpStruct->maxDiff5L);
	XtUnmanageChild(dpStruct->maxDiffL);
  }

 /* units L */
  if (strcmp(dpStruct->units,dpStruct->unitsString)) {
  strcpy(dpStruct->unitsString,dpStruct->units);
  str = XmStringCreateSimple(dpStruct->units);
  XtVaSetValues(dpStruct->unitsL,XmNlabelString,str,NULL);
  XmStringFree(str);
  }

 /* operRangeL */
  cvtDoubleToCompactString(dpStruct->lopr,loprString,
			(unsigned short)dpStruct->setPointPrecision);
  cvtDoubleToCompactString(dpStruct->hopr,hoprString,
			(unsigned short)dpStruct->setPointPrecision);
  sprintf(dpStruct->operRangeUnitsString,
	"[%s, %s] ",loprString,hoprString);
  nColumns = strlen(dpStruct->operRangeUnitsString);
  if (strcmp(dpStruct->operRangeUnitsString,dpStruct->operRangeString)) {
  strcpy(dpStruct->operRangeString,dpStruct->operRangeUnitsString);
  str = XmStringCreateSimple(dpStruct->operRangeUnitsString);
  XtVaSetValues(dpStruct->operRangeL,XmNlabelString,str,NULL);
  XmStringFree(str);
  }
  operRangeMaxColumns = MAX(nColumns,operRangeMaxColumns);


  /*
   *  update # columns in operRangeL based on width of string...
   *   if this new XmNcolumns is greater, increase everybody's so that
   *    the rows are all the same size, then right justify units...
   */
  if (nColumns > operRangeMaxColumns) {
    operRangeMaxColumns = nColumns;
    ele = listHead;
    XtSetArg(wargs[0],XmNwidth,10*operRangeMaxColumns);
    while (ele != NULL) {
      XtSetValues(ele->operRangeL,wargs,1);
      ele = ele->next;
    }
 /* may not really be necessary */
    XtVaGetValues(mainBB1,XmNwidth,&width,NULL);
    XtVaSetValues(mainShell,XmNwidth,width+MW_MARGIN,NULL);
  } else {
    XtVaSetValues(dpStruct->operRangeL,XmNwidth,10*operRangeMaxColumns,NULL);
  }

}



/***
 *** channel access routines
 ***/



/******************************************************************************
 *  This function initializes channel access
 ******************************************************************************/
void dpInitializeCA()
{

/*
 * add CA's fd to X
 */
        SEVCHK(ca_add_fd_registration(dpRegisterCA,NULL),
                "\ndpInitializeCA: error adding CA's fd to X");

/*
 * add CA heartbeat ca_pend_event();
 */
	XtAppAddTimeOut(appContext,10000,caHeartBeat,NULL);
}



/******************************************************************************
 *  This function terminates channel access
 *	and un-registers CA's fd with X... 
 ******************************************************************************/
void dpTerminateCA()
{

   /* cancel registration of the CA file descriptors */
      SEVCHK(ca_add_fd_registration(dpRegisterCA,NULL),
                "\ndpTerminateCA:  error removing CA's fd from X");
   /* and close channel access */
      ca_pend_event(20.0*CA_PEND_EVENT_TIME);	/* don't allow early returns */


      SEVCHK(ca_task_exit(),"\ndpTerminateCA: error exiting CA");

}


/******************************************************************************
 * functions to register Channel Access' file descriptor with Xt
 *      and perform CA handling when events exist on that event stream
 ******************************************************************************/

void dpRegisterCA(void *dummy, int fd, int condition)
{
    int currentNumInps;
    
#define NUM_INITIAL_FDS 10
    typedef struct {
	XtInputId inputId;
	int fd;
    } InputIdAndFd;
    static InputIdAndFd *inp = NULL;
    static int maxInps = 0, numInps = 0;
    int i, j, k;
    
    
    
    if (inp == NULL && maxInps == 0) { 
      /* first time through */
	inp = (InputIdAndFd *) calloc(1,NUM_INITIAL_FDS*sizeof(InputIdAndFd));
	maxInps = NUM_INITIAL_FDS;
	numInps = 0;
    }
    
    if (condition) {
      /*
       * add new fd
       */
	if (numInps < maxInps-1) {
	    
	    inp[numInps].fd = fd;
	    inp[numInps].inputId  = XtAppAddInput(appContext,fd,
	      (XtPointer)XtInputReadMask,
	      dpProcessCA,(XtPointer)NULL);
	    numInps++;
	    
	} else {

	    fprintf(stderr,"\ndpRegisterCA: info: realloc-ing input fd's array");
	    
	    maxInps = 2*maxInps;
	    inp = (InputIdAndFd *) realloc(inp,maxInps*sizeof(InputIdAndFd));
	    inp[numInps].fd = fd;
	    inp[numInps].inputId  = XtAppAddInput(appContext,fd,
	      (XtPointer)XtInputReadMask,
	      dpProcessCA,(XtPointer)NULL);
	    numInps++;
	}
	
    } else {
	
	currentNumInps = numInps;
	
      /*
       * remove old fd/inputId
       */
	for (i = 0; i < numInps; i++) {
	    if (inp[i].fd == fd) {
		XtRemoveInput(inp[i].inputId);
		inp[i].inputId = (XtInputId)NULL;
		inp[i].fd = (int)NULL; 
		currentNumInps--;
	    }
	}
	
      /* now remove holes in the array */
	i = 0;
	while (i < numInps) {
	    if (inp[i].inputId == (XtInputId)NULL) {
		j = i+1;
		k = 0;
		while(inp[j].inputId != (XtInputId)NULL) {
		    inp[i+k].inputId = inp[j].inputId;
		    inp[i+k].fd = inp[j].fd;
		    j++;
		    k++;
		}
		i = j-1;
	    }
	    i++;
	}
	numInps = currentNumInps;
	
    }
    
#ifdef DEBUG
    fprintf(stderr,"\ndpRegisterCA: numInps = %d\n\t",numInps);
    for (i = 0; i < maxInps; i++)
      fprintf(stderr,"%d ",inp[i].fd);
    fprintf(stderr,"\n");
#endif
    
}



/*
 * handle periodic update (1 second) of MaxDiff (5s)
 */
static void dpUpdateDiff5(XtPointer client_data, XtIntervalId *id)
{
  char string[80];
  int j, k;
  time_t currentTime;
  DpStruct *dpStruct;
  double newDiff5, newDiff5Magnitude;


/* code stolen from dpProcessCA() */

  dpStruct = listHead;

  while(dpStruct != (DpStruct *)NULL) {

    if (dpStruct->entryType == ProcessVariable ||
	(dpStruct->entryType == AtomicDevice && dpStruct->foundSetPoint)) {
/*
 * maxDiff5  - march through array round-robin:
 * e.g., 2,1,0,NUM_DIFFERNCES-1, NUM_DIFFERENCES-2,...
 */
      newDiff5 = dpStruct->readback - dpStruct->setPoint;
      if (newDiff5 != dpStruct->maxDiff5) {
        j = dpStruct->diffArrayIndex;
        k = 0;
        currentTime = time((time_t *)NULL);
        newDiff5Magnitude = fabs(newDiff5);
        while (k < NUM_DIFFERENCES &&
	  (currentTime - dpStruct->diffArray[j].time) <= TIME_FOR_DIFFERENCES) {
	  if (dpStruct->diffArray[j].diff > newDiff5Magnitude) {
	    newDiff5Magnitude = dpStruct->diffArray[j].diff;
	    newDiff5 = (dpStruct->diffArray[j].negative
		? -1.0 * dpStruct->diffArray[j].diff
		: dpStruct->diffArray[j].diff);
	  }
	  j--;
	  if (j < 0) j = NUM_DIFFERENCES + j;
	  k++;
        }

        dpStruct->maxDiff5 = newDiff5;
        cvtDoubleToCompactString(dpStruct->maxDiff5,string,
		(unsigned short)dpStruct->readbackPrecision);
       if (strcmp(string,dpStruct->maxDiff5String)) {
         strcpy(dpStruct->maxDiff5String,string);
        str = XmStringCreateSimple(string);
        XtVaSetValues(dpStruct->maxDiff5L,XmNlabelString,str,NULL);
        XmStringFree(str);
        }
      }
    }

    dpStruct = dpStruct->next;
  }

/* reregister 1 second TimeOut function to update maxDiff5 */
  intervalId = XtAppAddTimeOut(appContext,1000,dpUpdateDiff5,NULL);
}




static void dpProcessCA(XtPointer client_data, int *source, XtInputId *id)
{
  char string[80];
  int j, k;
  time_t currentTime;
  DpStruct *dpStruct;
  double newDiff5, newDiff5Magnitude;

  int firstDisplayedRow=0, lastDisplayedRow=0, displayedRowCount=0, rowNumber ;
  Dimension clipWindowHeight, rowHeight, marginHeightAroundRows=MARGIN ;
  int scrollbarValue;

  ca_pend_event(CA_PEND_EVENT_TIME);



/*
 * turns out that with even modestly changing channels, don't
 *  want one update per CA monitor event - hence back to old
 *  update-traversal mechanism
 *  (N.B. - more expensive calculations are done here just before display
 *  rather than at each monitor event)
 */

  dpStruct = listHead;

  rowNumber = 0;

  while (dpStruct != (DpStruct *)NULL) {

    if (dpStruct->modified ) {

/*
 * update only displayed rows
 */
      if (!displayedRowCount) {

        XtVaGetValues(clipWindow,XmNheight,&clipWindowHeight,NULL);
        XtVaGetValues(clipWindow,XmNmarginHeight,&marginHeightAroundRows,NULL);
        XtVaGetValues(verticalScrollBar,XmNvalue,&scrollbarValue,NULL);
        if (dpStruct->setPointTF != (Widget)NULL)
            XtVaGetValues(dpStruct->DA,XmNheight,&rowHeight,NULL);
        rowHeight=rowHeight-2;
  
        displayedRowCount =(int) (1+(int)(clipWindowHeight-2*marginHeightAroundRows)/
            (int)(rowHeight+MOVE_MARGIN));
        firstDisplayedRow = (int)((int)scrollbarValue/(int)(rowHeight+MOVE_MARGIN));
        lastDisplayedRow = firstDisplayedRow+(displayedRowCount-1);

      }

      if (rowNumber >= firstDisplayedRow && rowNumber <= lastDisplayedRow ) {

      dpStruct->modified = False;
/* setPoint */
      if (dpStruct->setPointTF != (Widget)NULL && dpStruct->setPointChanged) {
        cvtDoubleToCompactString(dpStruct->setPoint,string,
		(unsigned short)dpStruct->setPointPrecision);
       if (strcmp(string,dpStruct->setPointString)) {
         strcpy(dpStruct->setPointString,string);
       XmTextFieldSetString(dpStruct->setPointTF,string);
       }
	dpStruct->setPointChanged = False;
      }
/* readback */
      if (dpStruct->readbackL != (Widget)NULL && dpStruct->readbackChanged) {
        cvtDoubleToCompactString(dpStruct->readback,string,
		(unsigned short)dpStruct->readbackPrecision);
       if (strcmp(string,dpStruct->readbackString)) {
         strcpy(dpStruct->readbackString,string);
        str = XmStringCreateSimple(string);
        XtVaSetValues(dpStruct->readbackL,XmNlabelString,str,NULL);
        XmStringFree(str);
       }
	dpStruct->readbackChanged = False;
      }
/* average */
      if (dpStruct->averageL != (Widget)NULL) {
        cvtDoubleToCompactString(dpStruct->average,string,
		(unsigned short)dpStruct->readbackPrecision+2);
       if (strcmp(string,dpStruct->averageString)) {
         strcpy(dpStruct->averageString,string);
        str = XmStringCreateSimple(string);
        XtVaSetValues(dpStruct->averageL,XmNlabelString,str,NULL);
        XmStringFree(str);
        }
      }
/* rmsDiff */
      dpStruct->rmsDiff = sqrt(dpStruct->averageDiffSquared);
      if (dpStruct->rmsDiffL != (Widget)NULL) {
        cvtDoubleToCompactString(dpStruct->rmsDiff,string,
		(unsigned short)dpStruct->readbackPrecision+2);
       if (strcmp(string,dpStruct->rmsDiffString)) {
         strcpy(dpStruct->rmsDiffString,string);
        str = XmStringCreateSimple(string);
        XtVaSetValues(dpStruct->rmsDiffL,XmNlabelString,str,NULL);
        XmStringFree(str);
        }
      }
      if (dpStruct->maxDiff5L != (Widget)NULL) {
    /*
     * maxDiff5  - march through array round-robin:
     * e.g., 2,1,0,NUM_DIFFERNCES-1, NUM_DIFFERENCES-2,...
     */
         j = dpStruct->diffArrayIndex;
         k = 0;
         currentTime = time((time_t *)NULL);
         newDiff5 = dpStruct->readback - dpStruct->setPoint;
         newDiff5Magnitude = fabs(newDiff5);
         while (k < NUM_DIFFERENCES &&
	   (currentTime - dpStruct->diffArray[j].time) <= TIME_FOR_DIFFERENCES){
	   if (dpStruct->diffArray[j].diff > newDiff5Magnitude) {
	    newDiff5Magnitude = dpStruct->diffArray[j].diff;
	    newDiff5 = (dpStruct->diffArray[j].negative
		? -1.0 * dpStruct->diffArray[j].diff
		: dpStruct->diffArray[j].diff);
	   }
	   j--;
	   if (j < 0) j = NUM_DIFFERENCES + j;
	   k++;
         }

        dpStruct->maxDiff5 = newDiff5;
        cvtDoubleToCompactString(dpStruct->maxDiff5,string,
		(unsigned short)dpStruct->readbackPrecision);
        if (strcmp(string,dpStruct->maxDiff5String)) {
         strcpy(dpStruct->maxDiff5String,string);
        str = XmStringCreateSimple(string);
        XtVaSetValues(dpStruct->maxDiff5L,XmNlabelString,str,NULL);
        XmStringFree(str);
        }
      }

/* maxDiff */
      if (dpStruct->maxDiffL != (Widget)NULL && dpStruct->maxDiffChanged) {
        cvtDoubleToCompactString(dpStruct->maxDiff,string,
		(unsigned short)dpStruct->readbackPrecision);
        if (strcmp(string,dpStruct->maxDiffString)) {
         strcpy(dpStruct->maxDiffString,string);
        str = XmStringCreateSimple(string);
        XtVaSetValues(dpStruct->maxDiffL,XmNlabelString,str,NULL);
        XmStringFree(str);
        }
	dpStruct->maxDiffChanged = False;
      }

    }
    }
    dpStruct = dpStruct->next;
    rowNumber++;
  }
}






/******************************************************************************
 * Channel Access callback and event handling functions...
 ******************************************************************************/



/***
 ***  Monitor routines
 ***/


/*
 * monitor connection event handler - when the channel connects,
 *   register a GR type get_callback() to get the graphics
 *   information for the monitor object/widget
 */
void processReadbackConnectionEvent(struct connection_handler_args args)
{
  chtype channelType;

  DpStruct *dpStruct = (DpStruct *) ca_puser(args.chid);

/* (MDA) events can still be firing when we can't process them,
 *	thanks to asynchronicity
 * also do the right thing when an ioc disconnects/reboots
 */
  if (dpStruct == NULL) {
	fprintf(stderr,"\nprocessReadbackConnectionEvent: ca_puser = NULL");
	ca_clear_channel(args.chid);
	ca_pend_event(10.*CA_PEND_EVENT_TIME);
	return;
  }


  if (!intervalId) {
/* register 1 second TimeOut function to update maxDiff5 as soon as first
 *  channel has connected - this starts the process for all channels.
 */
    intervalId = XtAppAddTimeOut(appContext,1000,dpUpdateDiff5,NULL);
  }



  if (args.op != CA_OP_CONN_UP) {

  /* black out the current (readback) value if disconnected */
    str = XmStringCreateSimple("");
    strcpy(dpStruct->setPointString,"");
    XtVaSetValues(dpStruct->readbackL,XmNlabelString,str,NULL);
    XmStringFree(str);
    XtVaSetValues(dpStruct->setPointTF,XmNcursorPositionVisible,False,
			XmNvalue,"",XmNshadowThickness,0, NULL);


  } else {

    XtVaSetValues(dpStruct->setPointTF,XmNshadowThickness,2,
			XmNcursorPositionVisible,True, NULL);

    /* only add get callback and all that if never previously connected */

    if (!dpStruct->readbackPreviouslyConnected) {

      dpStruct->readbackPreviouslyConnected = True;
      switch(ca_field_type(args.chid)) {
	case DBF_STRING:
		channelType = DBR_GR_STRING;
		fprintf(stderr,
		  "\n processReadbackConnectionEvent: %s %s",
		  "invalid channel type: STRING for channel:  ",
		  ca_name(args.chid));
		break;
	case DBF_INT:
/*	case DBF_SHORT: -- INT and SHORT are same thing in db_access.h */
		channelType = DBR_GR_INT; break;
	case DBF_FLOAT:
		channelType = DBR_GR_FLOAT; break;
	case DBF_ENUM:
		channelType = DBR_GR_ENUM; break;
	case DBF_CHAR:
		channelType = DBR_GR_CHAR; break;
	case DBF_LONG:
		channelType = DBR_GR_LONG; break;
	case DBF_DOUBLE:
		channelType = DBR_GR_DOUBLE; break;
	default:
		fprintf(stderr,
		  "\nprocessReadbackConnectionEvent: unknown field type = %d",
		   ca_field_type(args.chid));
		return;
      }
      SEVCHK(ca_get_callback(channelType,args.chid,
	processReadbackGrGetCallback,NULL),
	"\nprocessReadbackConnectionEvent: error in ca_get_callback");
      ca_pend_event(CA_PEND_EVENT_TIME);
    }
  }
}



/*
 * monitor graphics get callback - when the graphics information is available,
 *   store it and register a STS type event handler
 */
void processReadbackGrGetCallback(struct event_handler_args args)
{
  DpStruct *dpStruct = ca_puser(args.chid);
  chtype stsChannelType;
  struct dbr_gr_int    *ciData;
  struct dbr_gr_float  *cfData;
  struct dbr_gr_enum   *ceData;
  struct dbr_gr_char   *ccData;
  struct dbr_gr_long   *clData;
  struct dbr_gr_double *cdData;

  char string[80];

/*
 * store status information in ChannelAccesData structure or widget
 *  note: set point defines units and operating range, unless no set point
 *  is defined, then these values come from readback...
 */
  switch(args.type) {
	case DBR_GR_INT:
/*	case DBR_GR_SHORT: -- INT and SHORT are same thing in db_access.h */
		ciData = (struct dbr_gr_int *) args.dbr;
		if (!dpStruct->foundSetPoint) {
		    strcpy(dpStruct->units,ciData->units);
		    dpStruct->lopr = (double) ciData->lower_disp_limit;
		    dpStruct->hopr = (double) ciData->upper_disp_limit;
		}
		dpStruct->readbackPrecision = 0;
		dpStruct->readback = (double) ciData->value;
		stsChannelType = DBR_STS_INT;
		break;
	case DBR_GR_FLOAT:
		cfData = (struct dbr_gr_float *) args.dbr;
		if (!dpStruct->foundSetPoint) {
		    strcpy(dpStruct->units,cfData->units);
		    dpStruct->lopr = (double) cfData->lower_disp_limit;
		    dpStruct->hopr = (double) cfData->upper_disp_limit;
		}
		dpStruct->readbackPrecision = cfData->precision;
		dpStruct->readback = (double) cfData->value;
		stsChannelType = DBR_STS_FLOAT;
		break;
	case DBR_GR_ENUM:
		ceData = (struct dbr_gr_enum *) args.dbr;
		if (!dpStruct->foundSetPoint) {
		    dpStruct->lopr = (double) 0.0;
		   /* high is number of enum values -1 */
		    dpStruct->hopr = (double) (ceData->no_str - 1.0);
		}
		dpStruct->readbackPrecision = 0;
		dpStruct->readback = (double) ceData->value;
		stsChannelType = DBR_STS_ENUM;
		break;
	case DBR_GR_CHAR:
		ccData = (struct dbr_gr_char *) args.dbr;
		if (!dpStruct->foundSetPoint) {
		    strcpy(dpStruct->units,ccData->units);
		    dpStruct->lopr = (double) ccData->lower_disp_limit;
		    dpStruct->hopr = (double) ccData->upper_disp_limit;
		}
		dpStruct->readbackPrecision = 0;
		dpStruct->readback = (double) ccData->value;
		stsChannelType = DBR_STS_CHAR;
		break;
	case DBR_GR_LONG:
		clData = (struct dbr_gr_long *) args.dbr;
		if (!dpStruct->foundSetPoint) {
		    strcpy(dpStruct->units,clData->units);
		    dpStruct->lopr = (double) clData->lower_disp_limit;
		    dpStruct->hopr = (double) clData->upper_disp_limit;
		}
		dpStruct->readbackPrecision = 0;
		dpStruct->readback = (double) clData->value;
		stsChannelType = DBR_STS_LONG;
		break;
	case DBR_GR_DOUBLE:
		cdData = (struct dbr_gr_double *) args.dbr;
		if (!dpStruct->foundSetPoint) {
		    strcpy(dpStruct->units,cdData->units);
		    dpStruct->lopr = (double) cdData->lower_disp_limit;
		    dpStruct->hopr = (double) cdData->upper_disp_limit;
		}
		dpStruct->readbackPrecision = cdData->precision;
		dpStruct->readback = (double) cdData->value;
		stsChannelType = DBR_STS_DOUBLE;
		break;

	default:
		fprintf(stderr,
		  "\nprocessReadbackGrGetCallback: unknown field type = %ld",
		   args.type);
		return;
  }


  dpStruct->modified = True;
  dpStruct->count = 1;


 /* readbackL */
  cvtDoubleToCompactString(dpStruct->readback,string,
    (unsigned short)dpStruct->readbackPrecision);
  if (strcmp(string,dpStruct->readbackString)) {
      strcpy(dpStruct->readbackString,string);
      str = XmStringCreateSimple(string);
      XtVaSetValues(dpStruct->readbackL,XmNlabelString,str,NULL);
      XmStringFree(str);
  }
  dpStruct->readbackChanged = False;


/*
 * if no set point defined, then update operating range, units
 */
  if (!dpStruct->foundSetPoint) updateOperatingRangeAndUnits(dpStruct);

/* if PV, set setPoint to be initial readback value */
  if (dpStruct->entryType == ProcessVariable)
	dpStruct->setPoint = dpStruct->readback;


/*
 * and register a STS type event handler
 */
  ca_add_event(stsChannelType, dpStruct->readbackChid,
	processReadbackMonitorEvent, NULL, &(dpStruct->readbackEvid));
  ca_pend_event(CA_PEND_EVENT_TIME);

}



/*
 * status event handler - when the channel changes, store the new value
 *   and alarm data and update the modified field in the 
 *   DpStruct structure
 */
void processReadbackMonitorEvent(struct event_handler_args args)
{
  DpStruct *dpStruct = ca_puser(args.chid);
  struct dbr_sts_int    *ciData;
  struct dbr_sts_float  *cfData;
  struct dbr_sts_enum   *ceData;
  struct dbr_sts_char   *ccData;
  struct dbr_sts_long   *clData;
  struct dbr_sts_double *cdData;


  if (dpStruct == NULL) return;

/* store value and alarm data in DpStruct structure */
  switch(args.type) {
	case DBR_STS_INT:
/*	case DBR_STS_SHORT: -- INT and SHORT are same thing in db_access.h */
		ciData = (struct dbr_sts_int *) args.dbr;
		dpStruct->readback = (double) ciData->value;
		break;
	case DBR_STS_FLOAT:
		cfData = (struct dbr_sts_float *) args.dbr;
		dpStruct->readback = (double) cfData->value;
		break;
	case DBR_STS_ENUM:
		ceData = (struct dbr_sts_enum *) args.dbr;
		dpStruct->readback = (double) ceData->value;
		break;
	case DBR_STS_CHAR:
		ccData = (struct dbr_sts_char *) args.dbr;
		dpStruct->readback = (double) ccData->value;
		break;
	case DBR_STS_LONG:
		clData = (struct dbr_sts_long *) args.dbr;
		dpStruct->readback = (double) clData->value;
		break;
	case DBR_STS_DOUBLE:
		cdData = (struct dbr_sts_double *) args.dbr;
		dpStruct->readback = (double) cdData->value;
		break;

	default:
		fprintf(stderr,
		  "\nprocessReadbackMonitorEvent: unknown field type = %ld",
			args.type);
		return;
  }

  dpStruct->modified = True;
  dpStruct->readbackChanged = True;

/*
 * update the dpStruct statistics
 */
  if (dpStruct->entryType != CompositeDevice) updateStatistics(dpStruct);

}

     


void processSetPointConnectionEvent(struct connection_handler_args args)
{
  chtype channelType;

  DpStruct *dpStruct = (DpStruct *) ca_puser(args.chid);

/* (MDA) events can still be firing when we can't process them,
 *	thanks to asynchronicity
 * also do the right thing when an ioc disconnects/reboots
 */
  if (dpStruct == NULL) {
	fprintf(stderr,"\nprocessSetPointConnectionEvent: ca_puser = NULL");
	ca_clear_channel(args.chid);
	ca_pend_event(10.*CA_PEND_EVENT_TIME);
	return;
  }


  if (!intervalId) {
/* register 1 second TimeOut function to update maxDiff5 as soon as first
 *  channel has connected - this starts the process for all channels.
 */
    intervalId = XtAppAddTimeOut(appContext,1000,dpUpdateDiff5,NULL);
  }



  if (args.op != CA_OP_CONN_UP) {

  /* black out the current (readback) value if disconnected */
    XmTextFieldSetString(dpStruct->setPointTF,"");
    XtVaSetValues(dpStruct->setPointTF,XmNcursorPositionVisible,False,
			XmNvalue,"",XmNshadowThickness,0, NULL);


  } else {

    XtVaSetValues(dpStruct->setPointTF,XmNshadowThickness,2,
			XmNcursorPositionVisible,True, NULL);

    /* only add get callback and all that if never previously connected */

    if (!dpStruct->setPointPreviouslyConnected) {

      dpStruct->setPointPreviouslyConnected = True;
      switch(ca_field_type(args.chid)) {
	case DBF_STRING:
		channelType = DBR_STS_STRING;
		fprintf(stderr,
		  "\n processSetPointConnectionEvent: %s %s",
		  "invalid channel type: STRING for channel:  ",
		  ca_name(args.chid));
		break;
	case DBF_INT:
/*	case DBF_SHORT: -- INT and SHORT are same thing in db_access.h */
		channelType = DBR_GR_INT; break;
	case DBF_FLOAT:
		channelType = DBR_GR_FLOAT; break;
	case DBF_ENUM:
		channelType = DBR_GR_ENUM; break;
	case DBF_CHAR:
		channelType = DBR_GR_CHAR; break;
	case DBF_LONG:
		channelType = DBR_GR_LONG; break;
	case DBF_DOUBLE:
		channelType = DBR_GR_DOUBLE; break;
	default:
		fprintf(stderr,
		  "\nprocessSetPointConnectionEvent: unknown field type = %d",
		   ca_field_type(args.chid));
		return;
      }
      SEVCHK(ca_get_callback(channelType,args.chid,
	processSetPointGrGetCallback,NULL),
	"\nprocessSetPointConnectionEvent: error in ca_get_callback");
      ca_pend_event(CA_PEND_EVENT_TIME);
    }
  }
}



/*
 * monitor graphics get callback - when the graphics information is available,
 *   store it and register a STS type event handler
 */
void processSetPointGrGetCallback(struct event_handler_args args)
{
  DpStruct *dpStruct = ca_puser(args.chid);
  chtype stsChannelType;
  struct dbr_gr_int    *ciData;
  struct dbr_gr_float  *cfData;
  struct dbr_gr_enum   *ceData;
  struct dbr_gr_char   *ccData;
  struct dbr_gr_long   *clData;
  struct dbr_gr_double *cdData;

  char string[80];

/*
 * store status information in ChannelAccesData structure or widget
 *  note: set point defines units and operating range, unless no set point
 *  is defined, then these values come from readback...
 */
  switch(args.type) {
	case DBR_GR_INT:
/*	case DBR_GR_SHORT: -- INT and SHORT are same thing in db_access.h */
		ciData = (struct dbr_gr_int *) args.dbr;
		strcpy(dpStruct->units,ciData->units);
		dpStruct->lopr = (double) ciData->lower_disp_limit;
		dpStruct->hopr = (double) ciData->upper_disp_limit;
		dpStruct->setPoint = (double) ciData->value;
		dpStruct->setPointPrecision = 0;
		stsChannelType = DBR_STS_INT;
		break;
	case DBR_GR_FLOAT:
		cfData = (struct dbr_gr_float *) args.dbr;
		strcpy(dpStruct->units,cfData->units);
		dpStruct->lopr = (double) cfData->lower_disp_limit;
		dpStruct->hopr = (double) cfData->upper_disp_limit;
		dpStruct->setPoint = (double) cfData->value;
		dpStruct->setPointPrecision = cfData->precision;
		stsChannelType = DBR_STS_FLOAT;
		break;
	case DBR_GR_ENUM:
		ceData = (struct dbr_gr_enum *) args.dbr;
		dpStruct->lopr = (double) 0.0;
	/* high is number of enum values -1 */
		dpStruct->hopr = (double) (ceData->no_str - 1.0);
		dpStruct->setPointPrecision = 0;
		dpStruct->setPoint = (double) ceData->value;
		stsChannelType = DBR_STS_ENUM;
		break;
	case DBR_GR_CHAR:
		ccData = (struct dbr_gr_char *) args.dbr;
		strcpy(dpStruct->units,ccData->units);
		dpStruct->lopr = (double) ccData->lower_disp_limit;
		dpStruct->hopr = (double) ccData->upper_disp_limit;
		dpStruct->setPointPrecision = 0;
		dpStruct->setPoint = (double) ccData->value;
		stsChannelType = DBR_STS_CHAR;
		break;
	case DBR_GR_LONG:
		clData = (struct dbr_gr_long *) args.dbr;
		strcpy(dpStruct->units,clData->units);
		dpStruct->lopr = (double) clData->lower_disp_limit;
		dpStruct->hopr = (double) clData->upper_disp_limit;
		dpStruct->setPointPrecision = 0;
		dpStruct->setPoint = (double) clData->value;
		stsChannelType = DBR_STS_LONG;
		break;
	case DBR_GR_DOUBLE:
		cdData = (struct dbr_gr_double *) args.dbr;
		strcpy(dpStruct->units,cdData->units);
		dpStruct->lopr = (double) cdData->lower_disp_limit;
		dpStruct->hopr = (double) cdData->upper_disp_limit;
		dpStruct->setPointPrecision = cdData->precision;
		dpStruct->setPoint = (double) cdData->value;
		stsChannelType = DBR_STS_DOUBLE;
		break;

	default:
		fprintf(stderr,
		  "\nprocessReadbackGrGetCallback: unknown field type = %ld",
		   args.type);
		return;
  }


  dpStruct->modified = True;
  dpStruct->count = 1;

 /* setPointTF */
  cvtDoubleToCompactString(dpStruct->setPoint,string,
			(unsigned short)dpStruct->setPointPrecision);
  XmTextFieldSetString(dpStruct->setPointTF,string);
  dpStruct->setPointChanged = True;

 /* update fixed fields (operating range and units) based on set point */
  updateOperatingRangeAndUnits(dpStruct);


/*
 * and register a STS type event handler if Atomic Device
 */
  if (dpStruct->entryType == AtomicDevice) {
    ca_add_event(stsChannelType, dpStruct->setPointChid,
	processSetPointMonitorEvent,NULL, &(dpStruct->setPointEvid));
    ca_pend_event(CA_PEND_EVENT_TIME);
  } else if (dpStruct->entryType == ProcessVariable) {
    dpStruct->setPoint = dpStruct->readback;	/* set point is reference */
  }

}




/*
 * status event handler - when the channel changes, store the new value
 *   and alarm data and update the modified field in the 
 *   DpStruct structure
 */
void processSetPointMonitorEvent(struct event_handler_args args)
{
  DpStruct *dpStruct = ca_puser(args.chid);
  struct dbr_sts_int    *ciData;
  struct dbr_sts_float  *cfData;
  struct dbr_sts_enum   *ceData;
  struct dbr_sts_char   *ccData;
  struct dbr_sts_long   *clData;
  struct dbr_sts_double *cdData;


/* store value and alarm data in DpStruct structure */
  switch(args.type) {
	case DBR_STS_INT:
/*	case DBR_STS_SHORT: -- INT and SHORT are same thing in db_access.h */
		ciData = (struct dbr_sts_int *) args.dbr;
		dpStruct->setPoint = (double) ciData->value;
		break;
	case DBR_STS_FLOAT:
		cfData = (struct dbr_sts_float *) args.dbr;
		dpStruct->setPoint = (double) cfData->value;
		break;
	case DBR_STS_ENUM:
		ceData = (struct dbr_sts_enum *) args.dbr;
		dpStruct->setPoint = (double) ceData->value;
		break;
	case DBR_STS_CHAR:
		ccData = (struct dbr_sts_char *) args.dbr;
		dpStruct->setPoint = (double) ccData->value;
		break;
	case DBR_STS_LONG:
		clData = (struct dbr_sts_long *) args.dbr;
		dpStruct->setPoint = (double) clData->value;
		break;
	case DBR_STS_DOUBLE:
		cdData = (struct dbr_sts_double *) args.dbr;
		dpStruct->setPoint= (double) cdData->value;
		break;

	default:
		fprintf(stderr,
		  "\nprocessSetPointMonitorEvent: unknown field type = %ld",
			args.type);
		return;
  }

  dpStruct->modified = True;
  dpStruct->setPointChanged = True;


/*
 * update the dpStruct statistics
 */
  if (dpStruct->entryType != CompositeDevice) updateStatistics(dpStruct);

}



/*
 * STATUS
 */
void processStatusConnectionEvent(struct connection_handler_args args)
{
  chtype channelType;

  DpStruct *dpStruct = (DpStruct *) ca_puser(args.chid);

/* (MDA) events can still be firing when we can't process them,
 *	thanks to asynchronicity
 * also do the right thing when an ioc disconnects/reboots
 */
  if (dpStruct == NULL) {
	fprintf(stderr,"\nprocessSetPointConnectionEvent: ca_puser = NULL");
	ca_clear_channel(args.chid);
	ca_pend_event(10.*CA_PEND_EVENT_TIME);
	return;
  }


  if (args.op != CA_OP_CONN_UP) {

  /* update current status indicator if disconnected */
    dpStruct->status = NoStatus;

  } else {

    /* only add get callback and all that if never previously connected */

    if (!dpStruct->statusPreviouslyConnected) {

      dpStruct->statusPreviouslyConnected = True;
      switch(ca_field_type(args.chid)) {
	case DBF_STRING:
		channelType = DBR_STS_STRING;
		fprintf(stderr,
		  "\n processStatusConnectionEvent: %s %s",
		  "invalid channel type: STRING for channel:  ",
		  ca_name(args.chid));
		break;
	case DBF_INT:
/*	case DBF_SHORT: -- INT and SHORT are same thing in db_access.h */
		channelType = DBR_STS_INT; break;
	case DBF_FLOAT:
		channelType = DBR_STS_FLOAT; break;
	case DBF_ENUM:
		channelType = DBR_STS_ENUM; break;
	case DBF_CHAR:
		channelType = DBR_STS_CHAR; break;
	case DBF_LONG:
		channelType = DBR_STS_LONG; break;
	case DBF_DOUBLE:
		channelType = DBR_STS_DOUBLE; break;
	default:
		fprintf(stderr,
		  "\nprocessStatusConnectionEvent: unknown field type = %d",
		   ca_field_type(args.chid));
		return;
      }
      SEVCHK(ca_get_callback(channelType,args.chid,
	processStatusGetCallback,NULL),
	"\nprocessSetPointConnectionEvent: error in ca_get_callback");
      ca_pend_event(CA_PEND_EVENT_TIME);
    }
  }
}

/*
 * monitor get callback - when the value information is available,
 *   store it and register a STS type event handler
 */
void processStatusGetCallback(struct event_handler_args args)
{
  DpStruct *dpStruct = ca_puser(args.chid);
  chtype stsChannelType;
  struct dbr_sts_int    *ciData;
  struct dbr_sts_float  *cfData;
  struct dbr_sts_enum   *ceData;
  struct dbr_sts_char   *ccData;
  struct dbr_sts_long   *clData;
  struct dbr_sts_double *cdData;
  double value;


/* store status information in ChannelAccesData structure or widget */
  switch(args.type) {
	case DBR_STS_INT:
/*	case DBR_STS_SHORT: -- INT and SHORT are same thing in db_access.h */
		ciData = (struct dbr_sts_int *) args.dbr;
		value = (double) ciData->value;
		stsChannelType = DBR_STS_INT;
		break;
	case DBR_STS_FLOAT:
		cfData = (struct dbr_sts_float *) args.dbr;
		value = (double) cfData->value;
		stsChannelType = DBR_STS_FLOAT;
		break;
	case DBR_STS_ENUM:
		ceData = (struct dbr_sts_enum *) args.dbr;
		value = (double) ceData->value;
		stsChannelType = DBR_STS_ENUM;
		break;
	case DBR_STS_CHAR:
		ccData = (struct dbr_sts_char *) args.dbr;
		value = (double) ccData->value;
		stsChannelType = DBR_STS_CHAR;
		break;
	case DBR_STS_LONG:
		clData = (struct dbr_sts_long *) args.dbr;
		value = (double) clData->value;
		stsChannelType = DBR_STS_LONG;
		break;
	case DBR_STS_DOUBLE:
		cdData = (struct dbr_sts_double *) args.dbr;
		value = (double) cdData->value;
		stsChannelType = DBR_STS_DOUBLE;
		break;

	default:
		fprintf(stderr,
		  "\nprocessReadbackGrGetCallback: unknown field type = %ld",
		   args.type);
		return;
  }

/*
 * per Claude's Device Access:  value of status PV = 0 -> GOOD, !0 -> BAD
 */
  if (value == 0.0)
    dpStruct->status = GoodStatus;
  else
    dpStruct->status = BadStatus;
 
  renderStatus(dpStruct);

/*
 * and register a STS type event handler
 */
  ca_add_event(stsChannelType, dpStruct->statusChid,
	processStatusMonitorEvent,NULL, &(dpStruct->statusEvid));
  ca_pend_event(CA_PEND_EVENT_TIME);

}

/*
 * status event handler - when the channel changes, store the new value
 */
void processStatusMonitorEvent(struct event_handler_args args)
{
  DpStruct *dpStruct = ca_puser(args.chid);
  struct dbr_sts_int    *ciData;
  struct dbr_sts_float  *cfData;
  struct dbr_sts_enum   *ceData;
  struct dbr_sts_char   *ccData;
  struct dbr_sts_long   *clData;
  struct dbr_sts_double *cdData;
  double value;


/* store value and alarm data in DpStruct structure */
  switch(args.type) {
	case DBR_STS_INT:
/*	case DBR_STS_SHORT: -- INT and SHORT are same thing in db_access.h */
		ciData = (struct dbr_sts_int *) args.dbr;
		value = (double) ciData->value;
		break;
	case DBR_STS_FLOAT:
		cfData = (struct dbr_sts_float *) args.dbr;
		value = (double) cfData->value;
		break;
	case DBR_STS_ENUM:
		ceData = (struct dbr_sts_enum *) args.dbr;
		value = (double) ceData->value;
		break;
	case DBR_STS_CHAR:
		ccData = (struct dbr_sts_char *) args.dbr;
		value = (double) ccData->value;
		break;
	case DBR_STS_LONG:
		clData = (struct dbr_sts_long *) args.dbr;
		value = (double) clData->value;
		break;
	case DBR_STS_DOUBLE:
		cdData = (struct dbr_sts_double *) args.dbr;
		value = (double) cdData->value;
		break;

	default:
		fprintf(stderr,
		  "\nprocessStatusMonitorEvent: unknown field type = %ld",
			args.type);
		return;
  }


/*
 * per Claude's Device Access:  value of status PV = 0 -> GOOD, !0 -> BAD
 */
  if (value == 0.0)
    dpStruct->status = GoodStatus;
  else
    dpStruct->status = BadStatus;
 
  renderStatus(dpStruct);

}




/*
 * update the statistics for the specified dpStruct
 */
void updateStatistics(DpStruct *dpStruct)
{
  double nMinusOneOverN, oneOverN, diffAbs, tmpDiff;

/* update the statistics */
  nMinusOneOverN = (double)(dpStruct->count-1)/(double)(dpStruct->count);
  oneOverN = 1.0/(double)(dpStruct->count);
  dpStruct->count = MIN(dpStruct->count+1,UINT_MAX);    /* increment */
  dpStruct->average = dpStruct->average*nMinusOneOverN +
				   dpStruct->readback*oneOverN;

  if (dpStruct->entryType != CompositeDevice) {

    tmpDiff = dpStruct->readback - dpStruct->setPoint;
    diffAbs = fabs(tmpDiff);
    dpStruct->averageDiffSquared =
	dpStruct->averageDiffSquared*nMinusOneOverN +
	(tmpDiff*tmpDiff)*oneOverN;
/* round-robin, use up to the NUM_DIFFERENCES slots */
    dpStruct->diffArrayIndex = (dpStruct->diffArrayIndex + 1) % NUM_DIFFERENCES;
    dpStruct->diffArray[dpStruct->diffArrayIndex].diff = diffAbs;
    if (tmpDiff < 0.0)
      dpStruct->diffArray[dpStruct->diffArrayIndex].negative = True;
    else
      dpStruct->diffArray[dpStruct->diffArrayIndex].negative = False;
    dpStruct->diffArray[dpStruct->diffArrayIndex].time = time((time_t *)NULL);

    if (diffAbs > dpStruct->maxDiffAbs) {
      dpStruct->maxDiffChanged = True;
      dpStruct->maxDiff = tmpDiff;
      dpStruct->maxDiffAbs = diffAbs;
    }
  }

}



/*
 * clear the statistics for the selected dpStructs
 */
void clearStatistics(void)
{
  DpStruct *dpStruct;
  int j;

  if (currentlySelectedDpStructs != (DpStruct **)NULL) {
/* clear the statistics for selected rows */
    j = 0;
    while (currentlySelectedDpStructs[j] != (DpStruct *)NULL) {
      dpStruct = currentlySelectedDpStructs[j];
      if (dpStruct->entryType != CompositeDevice) {
 /* only set modified if connected - this prevents clear problem on unconn. */
	if (dpStruct->foundSetPoint && dpStruct->setPointChid != NULL)
	  if (ca_state(dpStruct->setPointChid) == cs_conn)
		dpStruct->modified = True;
	if (dpStruct->readbackChid != NULL)
	  if (ca_state(dpStruct->readbackChid) == cs_conn)
		dpStruct->modified = True;

	dpStruct->count = 1;
	dpStruct->diffArrayIndex = -1;
	dpStruct->average = dpStruct->readback;
	dpStruct->averageDiffSquared = (double)0.0;
	dpStruct->maxDiff = (double)0.0;
	dpStruct->maxDiffAbs = (double)0.0;
	dpStruct->maxDiffChanged = True;
	dpStruct->setPointChanged = True;
 /* used for PV's where reference point is some initial value from CA
  *   or for Atomic Devices which don't have set point defined         */
	if (dpStruct->entryType == ProcessVariable ||
	   (dpStruct->entryType == AtomicDevice && !dpStruct->foundSetPoint))
	    dpStruct->setPoint = dpStruct->readback;
        updateStatistics(dpStruct);
      }
      j++;
    }
  }

}


/*
 * clear all statistics
 */
void clearAllStatistics(void)
{
  DpStruct *dpStruct;

/* clear the statistics */
  dpStruct = listHead;
  while (dpStruct != (DpStruct *)NULL) {
    if (dpStruct->entryType != CompositeDevice) {
 /* only set modified if connected - this prevents clear problem on unconn. */
      if (dpStruct->foundSetPoint && dpStruct->setPointChid != NULL)
	 if (ca_state(dpStruct->setPointChid) == cs_conn)
		dpStruct->modified = True;
      if (dpStruct->readbackChid != NULL)
	 if (ca_state(dpStruct->readbackChid) == cs_conn)
		dpStruct->modified = True;
      dpStruct->count = 1;
      dpStruct->diffArrayIndex = -1;
      dpStruct->average = dpStruct->readback;
      dpStruct->averageDiffSquared = (double)0.0;
      dpStruct->maxDiff = (double)0.0;
      dpStruct->maxDiffAbs = (double)0.0;
      dpStruct->maxDiffChanged = True;
      dpStruct->setPointChanged = True;
 /* used for PV's where reference point is some initial value from CA
  *   or for Atomic Devices which don't have set point defined         */
	if (dpStruct->entryType == ProcessVariable ||
	   (dpStruct->entryType == AtomicDevice && !dpStruct->foundSetPoint))
	    dpStruct->setPoint = dpStruct->readback;
      updateStatistics(dpStruct);
    }
    dpStruct = dpStruct->next;
  }

}
