/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/*
 *	Mark Anderson, Argonne National Laboratory:
 *		U.S. DOE, University of Chicago
 */


#include <stdio.h>
#include <time.h>
#include "dp.h"
#include "cvtFast.h"

#include <unistd.h>

#define PRINT_OK_BTN		0
#define PRINT_CANCEL_BTN	1

/* Function prototypes */
static void printHeading(FILE *stream, char *dateString, int *returnY,
  int pageCount);
static void printToStream(FILE *stream);
static void printToPrinter(char *printer);
static void printToFile(char *file);
static void printCallback(Widget w, XtPointer client_data,
  XtPointer call_data);
static void appendDpDateTimeFileName(char *bufferPtr);
static void simpleDestCallback(Widget w, XtPointer client_data,
  XtPointer call_data);
static void printerTextFieldCallback(Widget w, XtPointer client_data,
  XtPointer call_data);
Widget createPrintDialog(Widget parent);


static Widget printDestRB, printerLabel, printerTextField;
static Widget printFileLabel, printFileTextField;


/* use default PS units, coordinates: (landscape orientation)
 *   units = 1/72 inch, origin = lower left corner
 */
#define X0		0
#define Y0		0
#define X1		792
#define Y1		612

#define TITLE_PTS	9
#define BODY_PTS 	7

#define TITLE_X		30
#define TITLE_Y		40
#define V_SPACE		3
#define H_SPACE		6
#define STATUS_RADIUS	2

#define ROWS_PER_BAND	5	/* for ease of reading, use subtle fill areas */



static int X[N_LEGEND_ENTRIES];


static void printHeading(FILE *stream, char *dateString, int *returnY,
  int pageCount)
{
  int i, y, statusX, statusY;
  int counter, bandY, bandHeight;

  if (pageCount > 1) fprintf(stream,"\nshowpage");
  fprintf(stream,
      "\n0 setgray 0 setlinewidth 90 rotate 0 %d translate",-Y1);
/* date/time at top of page */
  fprintf(stream,"\nBodyF setfont");
  fprintf(stream,"\n%d %d moveto (DP   %s) RightShow",
			X1-TITLE_X, Y1-2*TITLE_Y/3,dateString);

/* frame box for legend */
  fprintf(stream,"\n0.80 setgray");
  fprintf(stream,"\nnewpath %d %d moveto",TITLE_X,Y1-TITLE_Y+TITLE_PTS);
  fprintf(stream,
	"\n%d %d lineto %d %d lineto %d %d lineto %d %d lineto closepath fill",
	X1-TITLE_X,Y1 - TITLE_Y + TITLE_PTS,
	X1-TITLE_X,Y1 - TITLE_Y - V_SPACE,
	TITLE_X,Y1 - TITLE_Y - V_SPACE,
	TITLE_X,Y1 - TITLE_Y + TITLE_PTS);

/* legend strings */
  fprintf(stream,"\n0 setgray");
  fprintf(stream,"\nTitleF setfont");
  y = Y1 - TITLE_Y;
  for (i = MIN_LEGEND_ENTRY; i <= MAX_LEGEND_ENTRY; i++) {
    fprintf(stream,"\n%d %d moveto (%s) show",X[i], y, legendString[i]);
  }

#ifdef INCLUDE_COLUMN_SEPARATORS
/* column separators */
  fprintf(stream,"\n0.98 setgray");
  for (i = 1; i < N_LEGEND_ENTRIES; i++) {
     x = X[i] - H_SPACE;
     fprintf(stream,"\nnewpath %d %d moveto %d %d lineto closepath stroke",
	x,y,x,Y0+TITLE_Y);
  }
#endif

/* update where non-heading text can begin */
  y -= BODY_PTS + V_SPACE;

  *returnY = y;


/* grey bands for rows */
  bandY = y + BODY_PTS;	/* since text origin is baseline */
  bandHeight = ROWS_PER_BAND*(BODY_PTS+V_SPACE);
  counter = 0;
  while (bandY > TITLE_Y) {
     if (counter % 2 == 1)
	fprintf(stream,"\n0.96 setgray");
     else
	fprintf(stream,"\n0.99 setgray");
     fprintf(stream,"\nnewpath %d %d moveto",TITLE_X,bandY);
     fprintf(stream,
	 "\n%d %d lineto %d %d lineto %d %d lineto %d %d lineto closepath fill",
	   X1-TITLE_X,bandY,
	   X1-TITLE_X,MAX(Y0+TITLE_Y, bandY-bandHeight),
	   TITLE_X,MAX(Y0+TITLE_Y, bandY-bandHeight),
	   TITLE_X,bandY);
     bandY -= bandHeight;
     counter++;
  }


/* page number at bottom of page */
  fprintf(stream,"\n0 setgray BodyF setfont");
  fprintf(stream,"\n%d %d moveto (Page %d) RightShow",
			X1-TITLE_X, 2*TITLE_Y/3, pageCount);

/* Status key/legend */
  statusX = TITLE_X+6*BODY_PTS;
  statusY = 2*TITLE_Y/3;
  fprintf(stream,"\n%d %d moveto (Status: ) show", TITLE_X, statusY);
  fprintf(stream,"\n%d %d moveto ( = Good) show", statusX, statusY);
  fprintf(stream,"\nnewpath %d %d %d 0 360 arc %s",
			statusX-STATUS_RADIUS,
			statusY+BODY_PTS-V_SPACE-STATUS_RADIUS,
			STATUS_RADIUS,"stroke");
  statusX = TITLE_X+12*BODY_PTS;
  fprintf(stream,"\n%d %d moveto ( = Bad) show", statusX, statusY);
  fprintf(stream,"\nnewpath %d %d %d 0 360 arc %s",
			statusX-STATUS_RADIUS,
			statusY+BODY_PTS-V_SPACE-STATUS_RADIUS,
			STATUS_RADIUS,"fill");

}


/*
 * main print routine which sets up header information, prints headers
 *  and actual data
 */
static void printToStream(FILE *stream)
{
  DpStruct *dpStruct;
  int i, len, y, pageCount, entryWidth;
  char buffer1[64], buffer2[64];
  time_t timeType;
  char *dateString;
  Boolean printIt;

  dpStruct = listHead;
  pageCount = 1;
/* heading */

  X[0] = TITLE_X;
  entryWidth = X1/(N_LEGEND_ENTRIES+1);
  for (i = 1; i <= MAX_LEGEND_ENTRY; i++) {
    X[i] = TITLE_X + 2*entryWidth/3 + i*entryWidth;
  }
  /* scoot oper range over since it is relatively larger and its left
   *  neighbor is units which is relatively smaller */
  X[OPER_RANGE] = X[OPER_RANGE] - entryWidth/3;

  timeType = time((time_t *)NULL);
  dateString = ctime(&timeType);

/* replace any '\n' with '\0' */
  len = strlen(dateString);
  for (i = 0; i < len; i++)
     if (dateString[i]=='\n') dateString[i] = '\0';

  fprintf(stream,"%%!PS-Adobe-1.0\n%%%%");
  fprintf(stream,"\n/TitleF /Helvetica findfont %d scalefont def",
			TITLE_PTS);
  fprintf(stream,"\n/BodyF /Helvetica findfont %d scalefont def",BODY_PTS);
  fprintf(stream,"\n/BodyFB /Helvetica-Bold findfont %d scalefont def",
			BODY_PTS);
  fprintf(stream,
	"\n/RightShow {dup stringwidth pop 0 exch sub 0 rmoveto show} def");


  printHeading(stream,dateString,&y,pageCount++);

  while (dpStruct != (DpStruct *)NULL) {

    switch (dpStruct->entryType) {
	case CompositeDevice:
		fprintf(stream,"\nBodyFB setfont");
		fprintf(stream,"\n%d %d moveto (%s) show",X[DEVICE],y,
			dpStruct->channel);
		fprintf(stream,"\nBodyF setfont");
		if (dpStruct->status != NoStatus) {
		  fprintf(stream,"\nnewpath %d %d %d 0 360 arc %s",
			X[SET_POINT]-H_SPACE-STATUS_RADIUS,
			y+BODY_PTS-V_SPACE-STATUS_RADIUS,
			STATUS_RADIUS,(dpStruct->status == GoodStatus ?
				"stroke" : "fill"));
		}
		y -= BODY_PTS + V_SPACE;
		if (y < TITLE_Y) printHeading(stream,dateString,&y,pageCount++);
		break;

	case AtomicDevice:
	case ProcessVariable:
		fprintf(stream,"\n%d %d moveto (%s) show",
			X[DEVICE] + (dpStruct->memberOfComposite ?
				H_SPACE : 0),y,dpStruct->channel);
		if (dpStruct->status != NoStatus) {
		  fprintf(stream,"\nnewpath %d %d %d 0 360 arc %s",
			X[SET_POINT]-H_SPACE-STATUS_RADIUS,
			y+BODY_PTS-V_SPACE-STATUS_RADIUS,
			STATUS_RADIUS,(dpStruct->status == GoodStatus ?
				"stroke" : "fill"));
		}
		printIt = False;
		if (dpStruct->entryType == ProcessVariable) {
		    if (dpStruct->readbackChid != NULL)
		        if (ca_state(dpStruct->readbackChid) != cs_never_conn)
			    printIt = True;
		} else if (dpStruct->entryType == AtomicDevice) {
		    if (dpStruct->setPointChid != NULL)
		        if (ca_state(dpStruct->setPointChid) == cs_conn)
			    printIt = True;
		}
		if (printIt) {
		  cvtDoubleToCompactString(dpStruct->setPoint,buffer1,
			(unsigned short)dpStruct->setPointPrecision);
		  fprintf(stream,"\n%d %d moveto (%s) show",X[SET_POINT],y,
			buffer1);
		}
		if (dpStruct->readbackChid != NULL) {
		  if (ca_state(dpStruct->readbackChid) == cs_conn) {
		    cvtDoubleToCompactString(dpStruct->readback,buffer1,
			(unsigned short)dpStruct->readbackPrecision);
		    fprintf(stream,"\n%d %d moveto (%s) show", X[RDBK],y,
			buffer1);
		  }
		  if (ca_state(dpStruct->readbackChid) != cs_never_conn) {
		    cvtDoubleToCompactString(dpStruct->average,buffer1,
			(unsigned short)dpStruct->readbackPrecision+2);
		    fprintf(stream,"\n%d %d moveto (%s) show",X[AVERAGE],y,
			buffer1);
		    cvtDoubleToCompactString(dpStruct->rmsDiff,buffer1,
			(unsigned short)dpStruct->readbackPrecision+2);
		    fprintf(stream,"\n%d %d moveto (%s) show",X[RMS],y,buffer1);
		    cvtDoubleToCompactString(dpStruct->maxDiff5,buffer1,
			(unsigned short)dpStruct->readbackPrecision);
		    fprintf(stream,"\n%d %d moveto (%s) show",
			X[DIFF5],y,buffer1);
		    cvtDoubleToCompactString(dpStruct->maxDiff,buffer1,
			(unsigned short)dpStruct->readbackPrecision);
		    fprintf(stream,"\n%d %d moveto (%s) show",X[DIFF],y,
			buffer1);

		    fprintf(stream,"\n%d %d moveto (%s) show",X[UNITS],y,
			dpStruct->units);

		    cvtDoubleToCompactString(dpStruct->lopr,buffer1,
			(unsigned short)dpStruct->setPointPrecision);
		    cvtDoubleToCompactString(dpStruct->hopr,buffer2,
			(unsigned short)dpStruct->setPointPrecision);
		    fprintf(stream,"\n%d %d moveto ([%s, %s]) show",
			X[OPER_RANGE],y,buffer1,buffer2);
		  }
		}
		y -= BODY_PTS + V_SPACE;
		if (y < TITLE_Y) printHeading(stream,dateString,&y,pageCount++);
		break;

	default:
		fprintf(stderr,"\nprintToStream: invalid dpStruct node");
		break;

    }
    dpStruct = dpStruct->next;
  }
  fprintf(stream,"\nshowpage\n");
}

#define PRINT_CMD "lp -d"

static void printToPrinter(char *printer)
{
  FILE *stream;
  char name[L_tmpnam+1];
  char *command;

  mkstemp(name);

  command = (char *)malloc((size_t)
	(strlen(PRINT_CMD)+strlen(printer)+strlen(name)+8));
  if (command != NULL) {
    stream = fopen(name,"w+");
    if (stream != (FILE *)NULL) {
      printToStream(stream);
      fclose(stream);
      sprintf(command,"%s%s %s",PRINT_CMD,printer,name);
      system(command);
      free(command);
/* (MDA) should remove the tmp file when safe to do so, but how do we know when?
      unlink(name);
*/
    } else {
      fprintf(stderr,"\nprintToPrinter: cannot open file %s for write",name);
    }
  } else {
    fprintf(stderr,"\nprintToPrinter: cannot malloc memory for printing!");
  }
}

static void printToFile(char *file)
{
  FILE *stream;

  stream = fopen(file,"w+");
  printToStream(stream);
  fclose(stream);
}


static void printCallback(Widget w, XtPointer client_data,
  XtPointer call_data)
{
  Boolean dontUnmanage;
  char *printer = NULL, *printFile = NULL;
  int btn = (intptr_t)client_data;
    
  dontUnmanage = False;

  switch(btn) {
    case PRINT_OK_BTN:
	if (XtIsSensitive(printerTextField)) {
	  printer = XmTextFieldGetString(printerTextField);
	  if (strlen(printer) == 0) {
	    dontUnmanage = True;
	    XBell(XtDisplay(printerTextField),100);
	    XSync(XtDisplay(printerTextField),False);
	    XBell(XtDisplay(printerTextField),100);
	    XSync(XtDisplay(printerTextField),False);
	  } else {
	    printToPrinter(printer);
            XtFree(printer);
	  }
	} else if (XtIsSensitive(printFileTextField)) {
	  printFile = XmTextFieldGetString(printFileTextField);
	  printToFile(printFile);
          XtFree(printFile);
	}
	if (!dontUnmanage) XtUnmanageChild(printDialog);
	break;

    case PRINT_CANCEL_BTN:
	XtUnmanageChild(printDialog);
	break;
  }
}


#define DATE_TIME_NAME_LENGTH	32	/* larger than MAX date-time name */

/*
 * put hour-min-sec-day-month-year_dp.ps into buffer
 */
static void appendDpDateTimeFileName(char *bufferPtr)
{
  time_t timeType;
  char *dateString;

  timeType = time((time_t *)NULL);
  dateString = ctime(&timeType);

/* file name of form: 010352_16Sep94_dp.ps for "01:03:52 16 Sep 94" */

/* ctime has a nicely consistent format which it puts things... */
  *bufferPtr++ = dateString[11];	/* hour (e.g., 01) */
  *bufferPtr++ = dateString[12];
  *bufferPtr++ = dateString[14];	/* minute (e.g., 03) */
  *bufferPtr++ = dateString[15];
  *bufferPtr++ = dateString[17];	/* second (e.g., 52) */
  *bufferPtr++ = dateString[18];
  *bufferPtr++ = '_';
  if (dateString[8] == ' ') *bufferPtr++ = '0';
  else *bufferPtr++ = dateString[8];		/* day (e.g., 16) */
  *bufferPtr++ = dateString[9];
  *bufferPtr++ = dateString[4];		/* month (e.g., Sep) */
  *bufferPtr++ = dateString[5];
  *bufferPtr++ = dateString[6];
  *bufferPtr++ = dateString[22];	/* year (e.g., 94) */
  *bufferPtr++ = dateString[23];

  *bufferPtr++ = '_';
  *bufferPtr++ = 'd';
  *bufferPtr++ = 'p';
  *bufferPtr++ = '.';
  *bufferPtr++ = 'p';
  *bufferPtr++ = 's';
  *bufferPtr++ = '\0';

}



#define PRINTER_BTN	0
#define FILE_BTN	1
/*
 * print dialog radiobox callback for specifying the destination
 */
static void simpleDestCallback(Widget w, XtPointer client_data,
  XtPointer call_data)
{
  int btn, len;
  char *textValue, *tail, *newBuffer;
  XmToggleButtonCallbackStruct *cbs=(XmToggleButtonCallbackStruct *)call_data;


/* do nothing for transitions to Unset, only handle Set */
  if (cbs->set == False) return;

  btn = (intptr_t) client_data;

  switch (btn) {
    case PRINTER_BTN:
/* make PrintFile.. insensitive */
	XtSetSensitive(printFileLabel,False);
	XtSetSensitive(printFileTextField,False);
/* make Printer.. sensitive */
	XtSetSensitive(printerLabel,True);
	XtSetSensitive(printerTextField,True);
	break;

    case FILE_BTN:
	textValue = XmTextFieldGetString(printFileTextField);
	len = strlen(textValue);
	tail = &(textValue[len-1]);
	while (*tail != '/' && len > 0) {
	  tail--;
	  len--;
	}
	tail++;
	*tail = '\0';
	newBuffer = (char *)(XtMalloc(len+DATE_TIME_NAME_LENGTH+1));
	strcpy(newBuffer,textValue);
	appendDpDateTimeFileName(&(newBuffer[len]));
	XmTextFieldSetString(printFileTextField,newBuffer);
	XtFree(textValue);
	XtFree(newBuffer);

/* make PrintFile.. sensitive */
	XtSetSensitive(printFileLabel,True);
	XtSetSensitive(printFileTextField,True);
/* make Printer.. insensitive */
	XtSetSensitive(printerLabel,False);
	XtSetSensitive(printerTextField,False);
	break;
  }
}


#define PRINTER_TEXT_FIELD	0
#define PRINT_FILE_TEXT_FIELD	1
/*
 * text field activate callback for printer and print file text fields
 */
static void printerTextFieldCallback(Widget w, XtPointer client_data,
  XtPointer call_data)
{
  int textField = (intptr_t) client_data;

/* could validate printers, etc, here */

  switch (textField) {
    case PRINTER_TEXT_FIELD:
	break;
    case PRINT_FILE_TEXT_FIELD:
	break;
  }
}


/*
 * create the Print dialog
 */
Widget createPrintDialog(Widget parent)
{
  int i, n;
  Arg args[12];
  Widget messageBox, form, mainLabel, mainSep, rbFrame, rbFrameLabel;
  XmString xmString;
  XmString buttons[4];
  XmButtonType buttonType[4];
  WidgetList children;
  Cardinal numChildren;
  Dimension labelHeight;
#define LOCAL_PATH_MAX	1023
  char buffer[LOCAL_PATH_MAX+60];
  char *bufferPtr;

  xmString = XmStringCreateSimple("Print Data...");
  n = 0;
  XtSetArg(args[n],XmNdialogTitle,xmString); n++;
  XtSetArg(args[n],XmNdialogType,XmDIALOG_TEMPLATE); n++;
  XtSetArg(args[n],XmNmwmDecorations,MWM_DECOR_ALL|MWM_DECOR_RESIZEH); n++;
  messageBox = XmCreateMessageDialog(parent,"msgBox",args,n);
/* in Motif1.2.2 can't set ok and cancel callbacks directly on MessageBox
   apparently, so do it manually */
  XtAddCallback(XmMessageBoxGetChild(messageBox,XmDIALOG_OK_BUTTON),
		XmNactivateCallback,printCallback,
		(XtPointer)PRINT_OK_BTN);
  XtAddCallback(XmMessageBoxGetChild(messageBox,XmDIALOG_CANCEL_BUTTON),
		XmNactivateCallback,printCallback,
		(XtPointer)PRINT_CANCEL_BTN);
  XmStringFree(xmString);

/* override message box's defaults here */
  n = 0;
  XtSetArg(args[n],XmNautoUnmanage,False); n++;
  XtSetValues(messageBox,args,n);

  n = 0;
  XtSetArg(args[n],XmNshadowThickness,0); n++;
  form = XmCreateForm(messageBox,"form",args,n);

  n = 0;
  rbFrame = XmCreateFrame(form,"rbFrame",args,n);

  xmString = XmStringCreateSimple(
	"Select Destination Printer or File for PostScript Output");
  n = 0;
  XtSetArg(args[n],XmNlabelString,xmString); n++;
  XtSetArg(args[n],XmNalignment,XmALIGNMENT_CENTER); n++;
  mainLabel = XmCreateLabel(form,"mainLabel",args,n);
  XmStringFree(xmString);
  XtManageChild(mainLabel);

  n = 0;
  XtSetArg(args[n],XmNorientation,XmHORIZONTAL); n++;
  mainSep = XmCreateSeparator(form,"mainSep",args,n);
  XtManageChild(mainSep);

  xmString = XmStringCreateSimple("Destination");
  n = 0;
  XtSetArg(args[n],XmNlabelString,xmString); n++;
  XtSetArg(args[n],XmNchildType,XmFRAME_TITLE_CHILD); n++;
  rbFrameLabel = XmCreateLabel(rbFrame,"rbFrameLabel",args,n);
  XmStringFree(xmString);
  XtManageChild(rbFrameLabel);
  XtVaGetValues(rbFrameLabel,XmNheight,&labelHeight,NULL);
  

  buttons[0] = XmStringCreateSimple("Printer");
  buttons[1] = XmStringCreateSimple("File");
  buttonType[0] = XmRADIOBUTTON;
  buttonType[1] = XmRADIOBUTTON;
  n = 0;
  XtSetArg(args[n],XmNbuttonCount,2); n++;
  XtSetArg(args[n],XmNbuttons,buttons); n++;
  XtSetArg(args[n],XmNbuttonType,buttonType); n++;
  XtSetArg(args[n],XmNbuttonSet,0); n++;	/* Initially to Printer */
  XtSetArg(args[n],XmNsimpleCallback,simpleDestCallback); n++;
  XtSetArg(args[n],XmNchildType,XmFRAME_WORKAREA_CHILD); n++;
  printDestRB = XmCreateSimpleRadioBox(rbFrame,"printDestRB",args,n);
  XmStringFree(buttons[0]);
  XmStringFree(buttons[1]);
/* now dig out the toggle button children and set shadow thicknesses to 0 */
  XtSetArg(args[0],XmNchildren,&children);
  XtSetArg(args[1],XmNnumChildren,&numChildren);
  XtGetValues(printDestRB,args,2);
  XtSetArg(args[0],XmNshadowThickness,0);
  for (i = 0; i < numChildren; i++) {
    XtSetValues(children[i],args,1);
  }
  XtManageChild(printDestRB);
  XtManageChild(rbFrame);


/* printer text entry field */
  xmString = XmStringCreateSimple("Printer:");
  n = 0;
  XtSetArg(args[n],XmNmarginLeft,16); n++;
  /* to vertically center with "normal" TF */
  XtSetArg(args[n],XmNmarginTop,4); n++;
  XtSetArg(args[n],XmNlabelString,xmString); n++;
  printerLabel= XmCreateLabel(form,"printerLabel",args,n);
  XmStringFree(xmString);
  XtManageChild(printerLabel);
  n = 0;
  XtSetArg(args[n],XmNcolumns,12); n++;
  XtSetArg(args[n],XmNvalue,getenv("PSPRINTER")); n++;
  printerTextField = XmCreateTextField(form,"printerTextField",args,n);
  XtAddCallback(printerTextField,XmNactivateCallback,
	printerTextFieldCallback,
	(XtPointer)PRINTER_TEXT_FIELD);
  XtManageChild(printerTextField);
  

/* print file (filename) text entry field */
  xmString = XmStringCreateSimple("Print File:");
  n = 0;
  XtSetArg(args[n],XmNlabelString,xmString); n++;
  /* to vertically center with "normal" TF */
  XtSetArg(args[n],XmNmarginTop,4); n++;
  printFileLabel= XmCreateLabel(form,"printFileLabel",args,n);
  XmStringFree(xmString);
  XtManageChild(printFileLabel);
  n = 0;
  XtSetArg(args[n],XmNcolumns,50); n++;
/* for some odd reason can't get PATH_MAX reliably defined between systems */
  buffer[0] = '\0';
  getcwd(buffer,LOCAL_PATH_MAX+1);
  bufferPtr = &(buffer[strlen(buffer)]);
  *bufferPtr++ = '/';

  appendDpDateTimeFileName(bufferPtr);

  XtSetArg(args[n],XmNvalue,buffer); n++;
  printFileTextField = XmCreateTextField(form,
	"printFileTextField",args,n);
  XtAddCallback(printFileTextField,XmNactivateCallback,
	printerTextFieldCallback,
	(XtPointer)PRINT_FILE_TEXT_FIELD);
  XtManageChild(printFileTextField);

/* since Destination originally set to Printer, make insensitive Print File.. */
  XtSetSensitive(printFileLabel,False);
  XtSetSensitive(printFileTextField,False);

/*
 * now set form attachments
 */
/* mainLabel */
  n = 0;
  XtSetArg(args[n],XmNtopAttachment,XmATTACH_FORM); n++;
  XtSetArg(args[n],XmNleftAttachment,XmATTACH_FORM); n++;
  XtSetArg(args[n],XmNrightAttachment,XmATTACH_FORM); n++;
  XtSetValues(mainLabel,args,n);
/* mainSep */
  n = 0;
  XtSetArg(args[n],XmNtopAttachment,XmATTACH_WIDGET); n++;
  XtSetArg(args[n],XmNtopWidget,mainLabel); n++;
  XtSetArg(args[n],XmNleftAttachment,XmATTACH_FORM); n++;
  XtSetArg(args[n],XmNrightAttachment,XmATTACH_FORM); n++;
  XtSetArg(args[n],XmNtopOffset,8); n++;	/* some arb. spacing */
  XtSetArg(args[n],XmNbottomOffset,8); n++;	/* some arb. spacing */
  XtSetValues(mainSep,args,n);
/* rbFrame */
  n = 0;
  XtSetArg(args[n],XmNtopAttachment,XmATTACH_WIDGET); n++;
  XtSetArg(args[n],XmNtopWidget,mainSep); n++;
  XtSetArg(args[n],XmNleftAttachment,XmATTACH_FORM); n++;
  XtSetArg(args[n],XmNbottomAttachment,XmATTACH_FORM); n++;
  XtSetArg(args[n],XmNtopOffset,16); n++;	/* some arb. spacing */
  XtSetArg(args[n],XmNbottomOffset,8); n++;	/* some arb. spacing */
  XtSetValues(rbFrame,args,n);
/* printerLabel */
  n = 0;
  XtSetArg(args[n],XmNtopAttachment,XmATTACH_OPPOSITE_WIDGET); n++;
  XtSetArg(args[n],XmNtopWidget,printDestRB); n++;
  XtSetArg(args[n],XmNtopOffset,labelHeight+3); n++;
  XtSetArg(args[n],XmNleftOffset,20); n++;
  XtSetArg(args[n],XmNleftAttachment,XmATTACH_WIDGET); n++;
  XtSetArg(args[n],XmNleftWidget,printDestRB); n++;
  XtSetValues(printerLabel,args,n);
/* printerTextField */
  n = 0;
  XtSetArg(args[n],XmNtopAttachment,XmATTACH_OPPOSITE_WIDGET); n++;
  XtSetArg(args[n],XmNtopWidget,printerLabel); n++;
  XtSetArg(args[n],XmNleftAttachment,XmATTACH_WIDGET); n++;
  XtSetArg(args[n],XmNleftWidget,printerLabel); n++;
  XtSetValues(printerTextField,args,n);
/* printFileLabel */
  n = 0;
  XtSetArg(args[n],XmNtopAttachment,XmATTACH_WIDGET); n++;
  XtSetArg(args[n],XmNtopOffset,4); n++;
  XtSetArg(args[n],XmNtopWidget,printerTextField); n++;
  XtSetArg(args[n],XmNrightAttachment,XmATTACH_OPPOSITE_WIDGET); n++;
  XtSetArg(args[n],XmNrightWidget,printerLabel); n++;
  XtSetValues(printFileLabel,args,n);
/* printFileTextField */
  n = 0;
  XtSetArg(args[n],XmNtopAttachment,XmATTACH_OPPOSITE_WIDGET); n++;
  XtSetArg(args[n],XmNtopWidget,printFileLabel); n++;
  XtSetArg(args[n],XmNleftAttachment,XmATTACH_WIDGET); n++;
  XtSetArg(args[n],XmNleftWidget,printFileLabel); n++;
  XtSetValues(printFileTextField,args,n);

  XtManageChild(form);
  return(messageBox);
}
